import { Router } from "express";
import * as _ from "lodash";
import * as jwt from "jsonwebtoken";
import * as bcrypt from "bcryptjs";
import {
  ControlsType,
  KeystrokeDataType,
  SignInPayload,
  SignUpPayload,
  TypedRequest,
  UpdatePasswordPayload,
  USER_TYPE,
} from "../config/types";
import Auth from "../models/Auth";
import Candidates from "../models/Candidate";
import Organizations from "../models/Organization";
import {
  processKeystrokeData,
  findUser,
  createSignupDataFromProcessedData,
  addAttemptToKeystrokeData,
  computeDataTendencies,
  processAttempt,
  updatePasswordFromProcessedData,
} from "../utilities/keystroke";
import {
  API_KEY,
  DEFAULT_CONTROLS,
  JWT_SECRET,
  SECRETE_KEY,
} from "../config/contants";
import { authMiddleware } from "../middlewares/auth";

const router = Router();

router.post("/signup", async (req: TypedRequest<{}, SignUpPayload>, res) => {
  const {
    passwords,
    keydown,
    keyup,
    email,
    userType,
    candidate,
    organization,
  } = req.body;
  const attemptCount = passwords.length;

  const auth = await Auth.findOne({ email });

  if (auth) {
    return res
      .status(403)
      .json({ success: false, msg: "User already exists!" });
  }

  const processedAttempts: KeystrokeDataType[] = Array(attemptCount)
    .fill(0)
    .map((v, i) =>
      processKeystrokeData({ keydown: keydown[i], keyup: keyup[i] })
    );

  const passwordsEqual = processedAttempts.every((v) =>
    _.isEqual(v.full.keys, processedAttempts[0].full.keys)
  );

  if (!passwordsEqual) {
    return res
      .status(403)
      .json({ success: false, msg: "Passwords don't match" });
  }

  try {
    let newUser;
    if (userType === USER_TYPE.CANDIDATE) {
      newUser = await Candidates.create(candidate);
    } else {
      newUser = await Organizations.create(organization);
    }

    const signupData = await createSignupDataFromProcessedData(
      email,
      passwords,
      userType,
      newUser.id,
      processedAttempts
    );

    await Auth.create(signupData);

    return res.json({
      success: true,
    });
  } catch (error) {
    return res
      .status(500)
      .json({ success: false, msg: "Error signing up", error });
  }
});

router.post("/login", async (req: TypedRequest<{}, SignInPayload>, res) => {
  const { password, keydown, keyup, email, userType } = req.body;

  const auth = await Auth.findOne({ email });

  if (!auth) {
    return res
      .status(403)
      .json({ success: false, msg: "User does not exist in DB" });
  }

  const processedAttempt: KeystrokeDataType = processKeystrokeData({
    keydown,
    keyup,
  });

  const isMatch = await bcrypt.compare(password, auth.password);

  if (!isMatch) {
    return res.status(403).json({ success: false, msg: "Invalid Credentials" });
  }

  const result = processAttempt({
    userKeystrokeData: auth.keystrokeData,
    attemptKeystrokeData: processedAttempt,
    controls: auth.controls || DEFAULT_CONTROLS,
  });

  if (result.accepted) {
    const newUserData = addAttemptToKeystrokeData({
      userData: auth,
      attemptKeystrokeData: processedAttempt,
    });

    await Auth.updateOne({ email }, { $set: newUserData });

    let user;

    if (userType === USER_TYPE.CANDIDATE) {
      user = await Candidates.findById(auth.userId);
    } else {
      user = await Organizations.findById(auth.userId);
    }

    const token = await jwt.sign({ userId: auth.userId }, JWT_SECRET, {
      expiresIn: "5h",
    });

    return res.json({
      success: true,
      user,
      token,
      keystrokeResult: {
        result,
        db: {
          hold: auth.keystrokeData.hold.means,
          flight: auth.keystrokeData.flight.means,
          dd: auth.keystrokeData.dd.means,
          full: auth.keystrokeData.full.means,
        },
        filteredDb: {
          hold: auth.keystrokeData.hold.filteredMeans,
          flight: auth.keystrokeData.flight.filteredMeans,
          dd: auth.keystrokeData.dd.filteredMeans,
          full: auth.keystrokeData.full.filteredMeans,
        },
        attempt: {
          hold: processedAttempt.hold.times,
          flight: processedAttempt.flight.times,
          dd: processedAttempt.dd.times,
          full: processedAttempt.full.times,
        },
        controls: auth.controls || DEFAULT_CONTROLS,
      },
    });
  } else {
    return res.status(403).json({ success: false, msg: "Invalid Credentials" });
  }
});

router.get("/tendencies/:userId", authMiddleware, async (req, res) => {
  findUser(req.params.userId)
    .then((_auth) => {
      if (_auth) {
        return res.json({
          db: _auth.keystrokeData,
          calc: computeDataTendencies(_auth.keystrokeData),
        });
      } else {
        throw new Error("User not found");
      }
    })
    .catch((error) => res.status(500).send(error));
});

router.post(
  "/update",
  authMiddleware,
  async (req: TypedRequest<{ userId: string }, ControlsType>, res) => {
    try {
      await Auth.updateOne(
        { userId: req.query.userId },
        { $set: { controls: req.body } }
      );
      return res.send("SUCCESS");
    } catch (error) {
      return res.status(400).send("FAILED");
    }
  }
);

router.post(
  "/change-pwd",
  authMiddleware,
  async (req: TypedRequest<{ userId: string }, UpdatePasswordPayload>, res) => {
    const { passwords, keydown, keyup, oldPassword } = req.body;
    const attemptCount = passwords.length;

    const auth = await Auth.findOne({ userId: req.query.userId });

    const isMatch = await bcrypt.compare(oldPassword, auth.password);

    if (!isMatch) {
      return res
        .status(403)
        .json({ success: false, msg: "Invalid Credentials" });
    }

    const processedAttempts: KeystrokeDataType[] = Array(attemptCount)
      .fill(0)
      .map((v, i) =>
        processKeystrokeData({ keydown: keydown[i], keyup: keyup[i] })
      );

    const passwordsEqual = processedAttempts.every((v) =>
      _.isEqual(v.full.keys, processedAttempts[0].full.keys)
    );

    if (!passwordsEqual) {
      return res
        .status(403)
        .json({ success: false, msg: "Passwords don't match" });
    }

    try {
      const newPasswordData = await updatePasswordFromProcessedData(
        passwords,
        processedAttempts
      );

      Auth.updateOne({ userId: req.query.userId }, { $set: newPasswordData });

      return res.json({
        success: true,
      });
    } catch (error) {
      return res
        .status(500)
        .json({ success: false, msg: "Error signing up", error });
    }
  }
);

router.get("/conference/key", authMiddleware, async (req, res) => {
  try {
    const payload = {
      apikey: API_KEY,
      permissions: [`allow_join`], // `ask_join` || `allow_mod`
      version: 2,
      roles: ["CRAWLER"],
    };

    const token = await jwt.sign(payload, SECRETE_KEY, {
      expiresIn: "5h",
      algorithm: "HS256",
    });

    return res.json({ token });
  } catch (error) {
    return res.status(400).send(error);
  }
});

module.exports = router;
